from typing import Dict, List, Tuple, Type
from enum import IntEnum

import pygame
from pymgp.math import Vector2
from pymgp.sceneManagement import Scene

import pymunk
import pymunk.pygame_util

from pymgp.settings import PhysicsSettings


class pymgp:
    """
    PyMGP.
    """


    __physicsSpace : pymunk.Space = None
    """
    Physics Space.
    """

    __fps : int = 60
    """
    Frame per Seconds.
    """

    __clock : pygame.time.Clock = None


    def __init__(self):
        self.__physicsSpace = None
        self.__fps = 60
        self.__clock = pygame.time.Clock()
        pass


    @staticmethod
    def init():
        """
        Init PyMGP.
        """
        # Init PyGame
        pygame.init()

        # Init Physics Engine
        pymgp.__physicsSpace = pymunk.Space()       
        pymgp.__physicsSpace.gravity = PhysicsSettings.gravity
        
        # Init clock
        pymgp.__fps = 60
        pymgp.__clock = pygame.time.Clock()
        pass

    
    @staticmethod
    def createWindow(title : str, width : int, height : int):
        # Create Native Window
        pygame.display.set_caption(title)
        screen = pygame.display.set_mode((width, height))

        # Setup FrameBuffer
        FrameBuffer.getInstance().setInteralSurface(screen)

        # Setup Gizmos
        Gizmos.init(screen)
        return screen


    @staticmethod
    def setFPS(fps : int = 60):
        pymgp.__fps = fps
        pass


    @staticmethod
    def getFPS() -> int:
        return pymgp.__fps

    
    @staticmethod
    def tick():
        pymgp.__clock.tick(pymgp.__fps)
        pass

    
    @staticmethod
    def getPhysicsSpace():
        return pymgp.__physicsSpace


    @staticmethod
    def getEvents():
        return pygame.event.get()


    @staticmethod
    def flip():
        pygame.display.flip()
        pass

    
    @staticmethod
    def quit():
        # Quit PyGame
        pygame.quit()

        # Clear
        FrameBuffer.getInstance().clear()
        ObjectManager.getInstance().clear()
        RigidbodyManager.getInstance().clear()
        MGPBehaviourManager.getInstance().clear()
        SpriteGroupManager.getInstance().clear()
        pass


class Gizmos:
    """
    Help to draw images in Scene or Game view.
    """


    __drawOptions = None
    """
    Draw options.
    """


    def __init__(self):
        self.__drawOptions = None
        pass


    @staticmethod
    def init(surface : pygame.Surface):
        Gizmos.__drawOptions = pymunk.pygame_util.DrawOptions(surface)
        pass


    @staticmethod
    def getDrawOptions():
        return Gizmos.__drawOptions


class FrameBuffer:
    """
    The FrameBuffer.
    """

    instance = None
    """
    An instance of FrameBuffer
    """


    __internalSurface : pygame.surface.Surface = None
    """
    Interal Surface.
    """


    def __init__(self):
        self.__internalSurface = None
        pass


    @staticmethod
    def getInstance():
        if FrameBuffer.instance == None:
            FrameBuffer.instance = FrameBuffer()
        return FrameBuffer.instance
    

    def setInteralSurface(self, surface : pygame.surface.Surface):
        self.__internalSurface = surface
        pass


    def getInternalSurface(self) -> pygame.surface.Surface:
        return self.__internalSurface

    
    def clear(self):
        self.__internalSurface = None
        pass


class Object:
    """
    The base class of all reference objects in pymgp.
    """

    
    name : str = None
    """Name of object"""
    
    
    def __init__(self, name : str):
        """
        Init for Object.

        Args: 
            name: Name of Object.
        """
        if name == None:
            name = self.__class__.__name__
        self.name = name

        ObjectManager.getInstance().addObject(self)
        pass

    
    def getInstanceID(self) -> int:
        """
        Get Instance ID
        
        Returns:
            (int) Object's InstanceID
        """
        return ObjectManager.getInstance().getInstanceIdOfObject(self)
    
    
    def toString(self) -> str:
        """
        Get Object Name
        
        Returns:
            (str) Object's Name
        """
        return self.name


    def __str__(self) -> str:
        return self.toString()

    
    @classmethod
    def instantiate(obj):
        """
        Clone original object and return it.

        Args:
            obj: Original object.

        Returns:
            A clone of original object.
        """

        # TODO
        return None


    @classmethod
    def destroy(obj):
        """
        Remove GameObject, Components and other resources.
        """

        # TODO
        pass


class Component(Object):
    """
    The base class of all the contents that attached to a GameObject.
    """


    gameObject = None
    """
    The GameObject that this component attached to.
    """


    def __init__(self, name: str = None):
        super().__init__(name)
        pass


    def awake(self):
        """
        Awake.
        """
        pass


    def addComponent(self, component):
        """
        Add a component to the GameObject.
        """
        return self.gameObject.addComponent(component)


    def compareTag(self, tag : str) -> bool:
        """
        Check if the GameObject is marked as tag.
        """
        return self.gameObject.compareTag(tag)

    
    def getComponent(self, componentClassName : str):
        """
        If the GameObject is attached a component which className is $componentClassName then returns it.
        """
        return self.gameObject.getComponent(componentClassName)


class GameObject(Object):
    """
    The base class of all the entities in a scene.
    """

    activeInHierarchy : bool = False
    """
    It defines whether the GameObject is active in a scene or not.
    """

    layer : int = -1
    """
    The Layer that the GameObject belongs to.
    """

    scene : Scene = None
    """
    The Scene that the GameObject belongs to.
    """

    tag : str = None
    """
    The Tag of the GameObject.
    """

    transform = None
    """
    The Transform component that attached to the GameObject.
    """

    __componentContainer : Dict = {}
    """
    A dictionary container that contains all the components that attached to the GameObject.

    Key: 
        The Class of the Component.

    Value:
        The Reference of the Component
    """

    __isActive : bool = True
    """
    Whether the GameObject is activated or disactivated.
    """

    def __init__(self, name = None):
        """
        Create a new GameObject named $name
        """
        super().__init__(name)
        self.__componentContainer = {}
        self.__isActive = True
        self.transform = Transform()

        # TODO
        pass

    def addComponent(self, component : Component) -> Component:
        """
        Add a component to the GameObject.
        """
        component.gameObject = self
        component.awake()
        if not self.__componentContainer.__contains__(component.__class__):
            self.__componentContainer[component.__class__] = component
        
        return component


    def compareTag(self, tag : str) -> bool:
        """
        Check if the GameObject is marked as tag.
        """
        return self.tag == tag

    
    def getComponent(self, componentClass : Type[Component]) -> Component:
        """
        If the GameObject is attached a component which class is $componentClass then returns it.
        """
        ret : Component = None
        if self.__componentContainer.__contains__(componentClass):
            ret = self.__componentContainer[componentClass]
        else:
            print(componentClass + " not Found.")
        return ret


    def setActive(self, value : bool):
        """
        Activate or Deactivate the GameObject.
        """
        self.__isActive = value
        pass
    

    @classmethod
    def find(name : str):
        """
        Find a GameObject by $name.
        """
        # TODO
        return None


class Transform(Component):
    """
    GameObject's position, rotation and scale.
    """

    childCount : int = 0
    """
    The child count that parent Transform has. 
    """


    position : Vector2 = None
    """
    World position.
    """


    rotation : float = 0.0
    """
    Rotation.
    """


    __parent = None
    """
    Parent Transform.
    """


    __children : List = []
    """
    Children of the Transform.
    """


    def __init__(self, name: str = None):
        super().__init__(name)
        self.__parent = None
        self.__children = []
        self.position = Vector2.zero()
        # TODO
        pass


    def detachChildren(self):
        # TODO
        pass


    def find(self, name : str):
        # TODO
        return None


    def getChild(self, index : int):
        # TODO
        return None

    
    def isChildOf(self, parent) -> bool:
        # TODO
        return False


    def lookAt(self, target : Vector2):
        # TODO
        pass


    def rotate(self, angle):
        # TODO
        pass


    def setParent(self, parent):
        self.__parent = parent
        pass


    def translate(self, translation : Vector2):
        # TODO
        pass


class Behaviour(Component):
    """
    Behaviour means a component that can be enabled or disabled.
    """

    enabled : bool = False
    """
    An enabled Behaviour can be updated, else cannot.
    """

    isActiveAndEnabled : bool = False
    """
    Whether the Behaviour is activated and enabled.
    """

    def __init__(self, name: str = None):
        super().__init__(name)
        pass


class MGPBehaviour(Behaviour):
    """
    The base class of all the scripts in pymgp.
    """

    def __init__(self, name: str = None):
        super().__init__(name)
        MGPBehaviourManager.getInstance().addBehaviour(self)
        pass

    
    def awake(self):
        """
        Callback when a behaviour awakes.
        """


    def start(self):
        """
        Callback when a behaviour starts.
        """
        pass


    def update(self):
        """
        Callback when a behaviour updates.
        """
        pass


    def onEnable(self):
        """
        Callback when a behaviour is enabled.
        """
        pass

    
    def onDisable(self):
        """
        Callback when a behaviour is disabled.
        """
        pass


class ObjectManager:
    """
    Manage all of the Objects.
    """

    instance = None
    """
    An instance of ObjectManager.
    """

    
    __instanceIdSeed : int = 0
    """
    The seed of instance id.
    """


    __objectDictionary  = {}
    """
    The object dictionary.
    Key:
        The reference of the object.
        
    Value:
        The instance id of the object.
    """


    def __init__(self) -> None:
        self.__objectDictionary = {}
        self.__instanceIdSeed = 0
        pass

    
    @staticmethod
    def getInstance():
        if ObjectManager.instance == None:
            ObjectManager.instance = ObjectManager()
        return ObjectManager.instance


    def clear(self):
        for obj in self.__objectDictionary.values():
            obj = None
        self.__objectDictionary.clear()
        self.__instanceIdSeed = 0
        pass

    
    def addObject(self, obj : Object):
        """
        Add an object to object dictionary.
        """
        self.__objectDictionary[obj] = self.__instanceIdSeed
        self.__instanceIdSeed += 1
        pass

    
    def getObject(self, instanceId : int) -> Object:
        """
        Get an object from object dictionary.
        """
        return self.__objectDictionary[instanceId]


    
    def getInstanceIdOfObject(self, obj : Object) -> int:
        """
        Get the instance id of an object.
        """
        if self.__objectDictionary.__contains__(obj):
            return self.__objectDictionary[obj]
        return -1


class MGPBehaviourManager:
    """
    Manage all of the MGPBehaviours.
    """


    instance = None
    """
    An instance of MGPBehaviourManager.
    """


    __behaviours = set()
    """
    All of the MGPBehaviours.
    """


    def __init__(self):
        self.__behaviours = set()
        pass


    @staticmethod
    def getInstance():
        if MGPBehaviourManager.instance == None:
            MGPBehaviourManager.instance = MGPBehaviourManager()
        return MGPBehaviourManager.instance


    def addBehaviour(self, behaviour : MGPBehaviour):
        if self.__behaviours.__contains__(behaviour):
            return
        self.__behaviours.add(behaviour)
        pass


    def removeBehaviour(self, behaviour : MGPBehaviour):
        if self.__behaviours.__contains__(behaviour):
            self.__behaviours.remove(behaviour)
        pass


    def awakeBehaviours(self):
        for behaviour in self.__behaviours:
            behaviour.awake()
        pass


    def startBehaviours(self):
        for behaviour in self.__behaviours:
            behaviour.start()
        pass


    def updateBehaviours(self):
        for behaviour in self.__behaviours:
            behaviour.update()
        pass


    def clear(self):
        self.__behaviours.clear()
        pass


class KeyCode(IntEnum):
    Alpha0            = pygame.K_0
    Alpha1            = pygame.K_1
    Alpha2            = pygame.K_2
    Alpha3            = pygame.K_3
    Alpha4            = pygame.K_4
    Alpha5            = pygame.K_5
    Alpha6            = pygame.K_6
    Alpha7            = pygame.K_7
    Alpha8            = pygame.K_8
    Alpha9            = pygame.K_9
    AC_BACK           = pygame.K_AC_BACK   
    AMPERSAND         = pygame.K_AMPERSAND   
    ASTERISK          = pygame.K_ASTERISK   
    AT                = pygame.K_AT   
    BACKQUOTE         = pygame.K_BACKQUOTE   
    BACKSLASH         = pygame.K_BACKSLASH   
    BACKSPACE         = pygame.K_BACKSPACE   
    BREAK             = pygame.K_BREAK   
    CAPSLOCK          = pygame.K_CAPSLOCK   
    CARET             = pygame.K_CARET   
    CLEAR             = pygame.K_CLEAR   
    COLON             = pygame.K_COLON   
    COMMA             = pygame.K_COMMA   
    CURRENCYSUBUNIT   = pygame.K_CURRENCYSUBUNIT   
    CURRENCYUNIT      = pygame.K_CURRENCYUNIT   
    DELETE            = pygame.K_DELETE   
    DOLLAR            = pygame.K_DOLLAR   
    DOWN              = pygame.K_DOWN   
    END               = pygame.K_END   
    EQUALS            = pygame.K_EQUALS   
    ESCAPE            = pygame.K_ESCAPE   
    EURO              = pygame.K_EURO   
    EXCLAIM           = pygame.K_EXCLAIM   
    F1                = pygame.K_F1   
    F10               = pygame.K_F10   
    F11               = pygame.K_F11   
    F12               = pygame.K_F12   
    F13               = pygame.K_F13   
    F14               = pygame.K_F14   
    F15               = pygame.K_F15   
    F2                = pygame.K_F2   
    F3                = pygame.K_F3   
    F4                = pygame.K_F4   
    F5                = pygame.K_F5   
    F6                = pygame.K_F6   
    F7                = pygame.K_F7   
    F8                = pygame.K_F8   
    F9                = pygame.K_F9   
    GREATER           = pygame.K_GREATER   
    HASH              = pygame.K_HASH   
    HELP              = pygame.K_HELP   
    HOME              = pygame.K_HOME   
    INSERT            = pygame.K_INSERT   
    KP0               = pygame.K_KP0   
    KP1               = pygame.K_KP1   
    KP2               = pygame.K_KP2   
    KP3               = pygame.K_KP3   
    KP4               = pygame.K_KP4   
    KP5               = pygame.K_KP5   
    KP6               = pygame.K_KP6   
    KP7               = pygame.K_KP7   
    KP8               = pygame.K_KP8   
    KP9               = pygame.K_KP9   
    KP_0              = pygame.K_KP_0   
    KP_1              = pygame.K_KP_1   
    KP_2              = pygame.K_KP_2   
    KP_3              = pygame.K_KP_3   
    KP_4              = pygame.K_KP_4   
    KP_5              = pygame.K_KP_5   
    KP_6              = pygame.K_KP_6   
    KP_7              = pygame.K_KP_7   
    KP_8              = pygame.K_KP_8   
    KP_9              = pygame.K_KP_9   
    KP_DIVIDE         = pygame.K_KP_DIVIDE   
    KP_ENTER          = pygame.K_KP_ENTER   
    KP_EQUALS         = pygame.K_KP_EQUALS   
    KP_MINUS          = pygame.K_KP_MINUS   
    KP_MULTIPLY       = pygame.K_KP_MULTIPLY   
    KP_PERIOD         = pygame.K_KP_PERIOD   
    KP_PLUS           = pygame.K_KP_PLUS   
    LALT              = pygame.K_LALT   
    LCTRL             = pygame.K_LCTRL   
    LEFT              = pygame.K_LEFT   
    LEFTBRACKET       = pygame.K_LEFTBRACKET   
    LEFTPAREN         = pygame.K_LEFTPAREN   
    LESS              = pygame.K_LESS   
    LGUI              = pygame.K_LGUI   
    LMETA             = pygame.K_LMETA   
    LSHIFT            = pygame.K_LSHIFT   
    LSUPER            = pygame.K_LSUPER   
    MENU              = pygame.K_MENU   
    MINUS             = pygame.K_MINUS   
    MODE              = pygame.K_MODE   
    NUMLOCK           = pygame.K_NUMLOCK   
    NUMLOCKCLEAR      = pygame.K_NUMLOCKCLEAR   
    PAGEDOWN          = pygame.K_PAGEDOWN   
    PAGEUP            = pygame.K_PAGEUP   
    PAUSE             = pygame.K_PAUSE   
    PERCENT           = pygame.K_PERCENT   
    PERIOD            = pygame.K_PERIOD   
    PLUS              = pygame.K_PLUS   
    POWER             = pygame.K_POWER   
    PRINT             = pygame.K_PRINT   
    PRINTSCREEN       = pygame.K_PRINTSCREEN   
    QUESTION          = pygame.K_QUESTION   
    QUOTE             = pygame.K_QUOTE   
    QUOTEDBL          = pygame.K_QUOTEDBL   
    RALT              = pygame.K_RALT   
    RCTRL             = pygame.K_RCTRL   
    RETURN            = pygame.K_RETURN   
    RGUI              = pygame.K_RGUI   
    RIGHT             = pygame.K_RIGHT   
    RIGHTBRACKET      = pygame.K_RIGHTBRACKET   
    RIGHTPAREN        = pygame.K_RIGHTPAREN   
    RMETA             = pygame.K_RMETA   
    RSHIFT            = pygame.K_RSHIFT   
    RSUPER            = pygame.K_RSUPER   
    SCROLLLOCK        = pygame.K_SCROLLLOCK   
    SCROLLOCK         = pygame.K_SCROLLOCK   
    SEMICOLON         = pygame.K_SEMICOLON   
    SLASH             = pygame.K_SLASH   
    SPACE             = pygame.K_SPACE   
    SYSREQ            = pygame.K_SYSREQ   
    TAB               = pygame.K_TAB   
    UNDERSCORE        = pygame.K_UNDERSCORE   
    UNKNOWN           = pygame.K_UNKNOWN   
    UP                = pygame.K_UP   
    A                 = pygame.K_a   
    B                 = pygame.K_b   
    C                 = pygame.K_c   
    D                 = pygame.K_d   
    E                 = pygame.K_e   
    F                 = pygame.K_f   
    G                 = pygame.K_g   
    H                 = pygame.K_h   
    I                 = pygame.K_i   
    J                 = pygame.K_j   
    K                 = pygame.K_k   
    L                 = pygame.K_l   
    M                 = pygame.K_m   
    N                 = pygame.K_n   
    O                 = pygame.K_o   
    P                 = pygame.K_p   
    Q                 = pygame.K_q   
    R                 = pygame.K_r  
    S                 = pygame.K_s   
    T                 = pygame.K_t   
    U                 = pygame.K_u   
    V                 = pygame.K_v   
    W                 = pygame.K_w   
    X                 = pygame.K_x   
    Y                 = pygame.K_y   
    Z                 = pygame.K_z   


class KeyEvent:
    """
    Store all the events about Key.
    """


    keyEvents : List[pygame.event.Event] = []
    """
    Key events.
    """
        

class Input:
    """
    Input helper. base on PyGame's input system
    """


    @staticmethod
    def getKey(keyCode : KeyCode) -> bool:
        keyPressed = pygame.key.get_pressed()
        return keyPressed[keyCode]


    @staticmethod
    def getKeyDown(keyCode : KeyCode) -> bool:
        for keyEvent in KeyEvent.keyEvents:
            if keyEvent.type == pygame.KEYDOWN:
                if keyEvent.key == keyCode:
                    return True
        return False


    @staticmethod
    def getKeyUp(keyCode : KeyCode) -> bool:
        for keyEvent in KeyEvent.keyEvents:
            if keyEvent.type == pygame.KEYUP:
                if keyEvent.key == keyCode:
                    return True
        return False


class Renderer(Component):
    """
    Basic component for Rendering Objects.
    """

    sortingLayerID : int = 0
    """
    The unique Sorting Layer's ID.
    """

    sortingOrder : int = 0
    """
    The sorting order.
    """

    def __init__(self, name: str = None):
        super().__init__(name)
        RenderList.getInstance().append(self)
        pass


    def doSetup(self):
        """
        Setup something.
        """
        pass


    def doRendering(self):
        """
        Render someting.
        """
        pass


class RenderList:
    """
    A list stores all of the Renderers.
    """

    instance = None
    """
    An instance of RenderList
    """

    __internalList : List = []
    """
    Interal Render List.
    """


    def __init__(self):
        self.__internalList = []
        pass


    @staticmethod
    def getInstance():
        if RenderList.instance == None:
            RenderList.instance = RenderList()
        return RenderList.instance


    def append(self, renderer : Renderer):
        self.__internalList.append(renderer)
        pass


    def remove(self, renderer : Renderer):
        self.__internalList.remove(renderer)
        pass


    def doRendering(self):
        """
        Render all.
        """
        for renderer in self.__internalList:
            renderer.doRendering()
        pass


class SpriteGroupManager:
    """
    Manage all of the SpriteGroups.
    """


    instance = None
    """
    An instance of SpriteGroupManager.
    """


    __spriteGroups = {}
    """
    Sprite Group Dictionary.
    """


    def __init__(self) -> None:
        # Add a default SpriteGroup
        self.__spriteGroups["Default"] = pygame.sprite.Group()
        pass


    @staticmethod
    def getInstance():
        if SpriteGroupManager.instance == None:
            SpriteGroupManager.instance = SpriteGroupManager()
        return SpriteGroupManager.instance


    def addGroup(self, groupName : str, group : pygame.sprite.Group = None):
        if group == None:
            self.__spriteGroups[groupName] = pygame.sprite.Group()
        
        self.__spriteGroups[groupName] = group
        pass


    def getGroup(self, groupName : str) -> pygame.sprite.Group:
        return self.__spriteGroups[groupName]

    
    def clear(self):
        for group in self.__spriteGroups.values():
            group.empty()
        pass


class Sprite(Object):
    """
    Sprite.
    """


    __internalSprite : pygame.sprite.Sprite = None
    """
    PyGame's Sprite.
    """


    def __init__(self, path: str, name: str = None):
        super().__init__(name)
        self.__internalSprite = pygame.sprite.Sprite()
        self.__internalSprite.image = pygame.image.load(path)
        self.__internalSprite.rect = self.__internalSprite.image.get_rect()


    def getInternalSprite(self) -> pygame.sprite.Sprite:
        return self.__internalSprite


    def setPosition(self, rectCenter : Tuple):
        self.__internalSprite.rect.center = rectCenter

        
class SpriteRenderer(Renderer):
    """
    Renderer a Sprite.
    """

    sprite : Sprite = None
    """
    Sprite.
    """


    spriteGroupName : str = None
    """
    Sprite's Group name.
    """


    def __init__(self, name: str = None):
        super().__init__(name)
        
        pass


    def doSetup(self, sprite : Sprite, spriteGroupName : str):
        self.sprite = sprite
        self.spriteGroupName = spriteGroupName

        SpriteGroupManager.getInstance().getGroup(spriteGroupName).add(sprite.getInternalSprite())
        pass
    

    def doRendering(self):
        # Update Sprite's position.
        position = self.gameObject.transform.position
        self.sprite.setPosition(position.getTuple())

        # Update and Draw.
        SpriteGroupManager.getInstance().getGroup(self.spriteGroupName).update()
        surface = FrameBuffer.getInstance().getInternalSurface()
        SpriteGroupManager.getInstance().getGroup(self.spriteGroupName).draw(surface)
        pass


class Rigidbody2DType(IntEnum):
    DYNAMIC         = pymunk.Body.DYNAMIC
    STATIC          = pymunk.Body.STATIC
    KINEMATIC       = pymunk.Body.KINEMATIC


class Rigidbody2D(Component):
    """
    2D Rigidbody component.
    """


    __internalBody : pymunk.Body = None
    """
    Internal Rigidbody.
    """


    @property
    def mass(self) -> float:
        return self.__internalBody.mass


    @mass.setter
    def mass(self, value : float):
        self.__internalBody.mass = value
        pass


    @property
    def position(self) -> Tuple[int, int]:
        return self.__internalBody.position
    

    @position.setter
    def position(self, value : Tuple[int, int]):
        self.__internalBody.position = value
        pass


    def __init__(self, bodyType : Rigidbody2DType = Rigidbody2DType.DYNAMIC, name: str = None):
        super().__init__(name)
        self.__internalBody = pymunk.Body(1, 0, bodyType)
        RigidbodyManager.getInstance().addRigidbody(self)
        pass

    
    def awake(self):
        super().awake()
        self.__internalBody.position = self.gameObject.transform.position.getTuple()
        pass


    def update(self):
        self.gameObject.transform.position = Vector2(self.__internalBody.position.x, self.__internalBody.position.y)
        pass


    def getInternalBody(self) -> pymunk.Body:
        return self.__internalBody

    
    def addForce(self, force : Vector2):
        self.__internalBody.apply_impulse_at_local_point(force.getTuple())


class Collider2D(Behaviour):
    """
    2D Colliders' base class.
    """


    attachedRigidbody : Rigidbody2D = None
    """
    Attached Rigidbogy2D component.
    """


    __internalShape : pymunk.Shape = None
    """
    Internal Shape.
    """


    def awake(self):
        self.attachedRigidbody = self.gameObject.getComponent(Rigidbody2D)
        pass


class CircleCollider2D(Collider2D):
    """
    2D Circle Collider.
    """


    radius : float = 1.0
    """
    Radius of shape
    """


    offset : Tuple[float, float] = (0, 0)
    """
    Local offset of shape's position.
    """


    def __init__(self, radius, offset, name: str = None):
        super().__init__(name)
        self.radius = radius
        self.offset = offset
        pass


    def awake(self):
        super().awake()
        body = self.attachedRigidbody.getInternalBody()
        body.moment = pymunk.moment_for_circle(body.mass, 0, self.radius, (0, 0))
        self.__internalShape = pymunk.Circle(body, self.radius, self.offset)
        pymgp.getPhysicsSpace().add(body, self.__internalShape)
        pass


class BoxCollider2D(Collider2D):
    """
    2D Box Collider.
    """


    width : float = 0.0
    """
    Width of the Box.
    """


    height : float = 0.0
    """
    Height of the Box.
    """


    def __init__(self, width, height, name: str = None):
        super().__init__(name)
        self.width = width
        self.height = height
        pass


    def awake(self):
        super().awake()
        body = self.attachedRigidbody.getInternalBody()
        body.moment = pymunk.moment_for_box(body.mass, (self.width, self.height))
        self.__internalShape = pymunk.Poly.create_box(body, (self.width, self.height))
        pymgp.getPhysicsSpace().add(body, self.__internalShape)
        pass


class RigidbodyManager:
    """
    Manage all of the Rigidbodies.
    """


    instance = None
    """
    An instance of RigidbodyManager
    """


    rigidbodySet = set()
    """
    A set stores all of the rigidbodies.
    """


    @staticmethod
    def getInstance():
        if RigidbodyManager.instance == None:
            RigidbodyManager.instance = RigidbodyManager()
        return RigidbodyManager.instance

    
    def addRigidbody(self, body : Rigidbody2D):
        self.rigidbodySet.add(body)
        pass


    def updateRigidbodies(self):
        for body in self.rigidbodySet:
            body.update()
        pass


    def clear(self):
        self.rigidbodySet.clear()
        pass













